#[inline(always)]
fn print_printable_or_dot(chunk: &[u8]) {
    for (idx, byte) in chunk.iter().enumerate() {
        if idx > 0 && idx % 8 == 0 {
            print!(" ");
        }

        if *byte >= 32 && *byte <= 126 {
            print!("{}", String::from_utf8(vec![*byte]).unwrap());
        } else {
            print!(".");
        }
    }
}

#[inline(always)]
fn print_offset(offset: usize) {
    print!("{offset:08x}  ");
}

#[inline(always)]
fn print_data_hex(chunk: &[u8]) {
    for (idx, byte) in chunk.iter().enumerate() {
        if idx > 0 && idx % 8 == 0 {
            print!("  ");
        }
        print!("{byte:02x} ");
    }
}

#[inline(always)]
fn print_padding(chunk_len: usize) {
    // only needs padding if we have less than 16 bytes to show
    if chunk_len >= 16 {
        return;
    }

    // padding is 3 spaces per missing byte
    for _ in 0..(16 - chunk_len) {
        print!("   ");
    }

    // if 8 or more bytes are missing we also need pad the space between
    // 8-bytes left and 8-bytes right
    if chunk_len <= 8 {
        print!("  ");
    }
}

pub fn print_hex(data: &[u8]) {
    let mut offset = 0usize;

    let chunks: Vec<&[u8]> = data.chunks(16).collect();

    for chunk in chunks {
        print_offset(offset);
        print_data_hex(chunk);
        print_padding(chunk.len());
        print_printable_or_dot(chunk);
        println!();
        offset += 16;
    }
}
